#!/usr/bin/env python
import sys
import os
import logging
import textwrap
from typing import Type, Tuple
from pathlib import Path
from json import load
from datetime import datetime
from email.message import EmailMessage
from socket import gethostname
from email.utils import parseaddr
from json import dump

from dateutil.parser import parse
from pytz import UTC
from pydantic import (
    BaseModel,
    EmailStr,
    Field,
    ValidationError,
    AwareDatetime,
)
from pydantic_settings import (
    BaseSettings,
    SettingsConfigDict,
    YamlConfigSettingsSource,
    PydanticBaseSettingsSource,
)
from pydantic.json import pydantic_encoder
from aiosmtplib import SMTPRecipientsRefused, send
from aiosmtplib.errors import SMTPException, SMTPAuthenticationError

logging.basicConfig(
    level=logging.INFO,
    format="%(levelname)-8s - %(message)s",  # "{levelnames:10}{message}",
)
logger = logging.getLogger(__name__)

CONFIG = os.environ.get("CONFIG")
USER = os.environ.get("USER", "root")
HOST_NAME = gethostname()


class MailSettings(BaseSettings):
    """Definition of the email"""

    model_config = SettingsConfigDict(env_prefix="MAIL_")
    sender: str = Field(description="Email address from", default=f"{USER}@{HOST_NAME}")
    subject: str = Field(description="Subject (title) of the mail", default="Notice")
    ## Note: variables from User list in the body template
    template: str = Field(
        description="Template for the mail body, in new Python string template style,"
        "Valid fields are from user",
        default=textwrap.dedent(
            """\
        Dear {username},
        Just to say hello.
        Regards
        """
        ),
    )


class SourceSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="SOURCE_")
    json_file: Path | None = Field(
        description="JSON file with a list of records, with at least the keys: username, email",
        default=None,
    )
    to: list[str] = Field(
        description='Mails in the mailbox format("John Doe"<j.doe@example.com>',
        default=[],
    )
    online_min: datetime = Field(
        description="Filter records with a lastonline record",
        default=parse("1900-01-01"),
    )
    online_max: datetime = Field(
        description="Filter records with a lastonline record",
        default=parse("3000-01-01"),
    )


class ServerSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="SERVER_")
    hostname: str = "localhost"
    port: int = 25
    username: str | None = None
    password: str | None = None
    start_tls: bool = False
    use_tls: bool = False


class Settings(BaseSettings):
    model_config = SettingsConfigDict(yaml_file=CONFIG)
    mail: MailSettings = MailSettings()
    source: SourceSettings = SourceSettings()
    server: ServerSettings = ServerSettings()
    results_dir: Path = Field(
        description="Base directory to store results", default=Path.cwd() / "results"
    )
    redo_all: bool = Field(
        description="Bypass the results of the last attempt: try to resend all",
        default=False,
    )
    stop_on_auth_error: bool = Field(
        description="Stop any processing if the server replies with auth failure",
        default=True,
    )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        return (YamlConfigSettingsSource(settings_cls),)


settings = Settings()


class User(BaseModel):
    username: str
    email: EmailStr
    lastonline: AwareDatetime
    status: str
    uid: int
    confirmed: int = Field(alias="email:confirmed", default=0)
    failure: str | None = None


def read_json(json_file: Path) -> list[User]:
    with open(json_file) as f:
        users_full = load(f)
        # return Users.parse_obj(users_full)
        users = []
        for user in users_full:
            try:
                users.append(User(**user))
            except ValidationError:
                logger.info(f"User {user} is not valid")
        return users


class ProcessResults(BaseModel):
    success: list[User] = []
    failures: list[User] = []


async def process_users(users: list[User]) -> ProcessResults:
    results = ProcessResults()
    for user in users:
        try:
            await send_mail(user)
        except SMTPException as err:
            if isinstance(err, SMTPRecipientsRefused):
                message = err.recipients[0].message
            else:
                message = err.message
            logger.warning(f"Cannot send mail to {user.email}: {message}")
            user.failure = message
            results.failures.append(user)
            if isinstance(err, SMTPAuthenticationError) and settings.stop_on_auth_error:
                exit(1)
        else:
            logger.info(f"Mail sent to {user.email}")
            results.success.append(user)
    return results


async def send_mail(user: User):
    msg = EmailMessage()
    msg["Subject"] = settings.mail.subject
    msg["From"] = settings.mail.sender
    msg["To"] = user.email
    fmt = dict(user)
    msg.set_content(settings.mail.template.format_map(fmt))
    await send(
        msg,
        hostname=settings.server.hostname,
        username=settings.server.username,
        password=settings.server.password,
        port=settings.server.port,
        start_tls=settings.server.start_tls,
        use_tls=settings.server.use_tls,
    )


async def mailer():
    assert CONFIG is not None
    if not Path(CONFIG).is_file():
        logger.warning(f"No {CONFIG} file, aborting.")
        sys.exit(1)
    logging.info(f"Using {CONFIG}")
    now = datetime.now().replace(microsecond=0)
    result_dir = settings.results_dir / Path(CONFIG).stem
    result_dir.mkdir(parents=True, exist_ok=True)

    failure_files = [f for f in result_dir.glob("*.err")]
    if len(failure_files) > 0:
        last_failure = sorted(failure_files, key=lambda f: f.lstat().st_mtime)[-1]
        users = read_json(last_failure)
        if len(users) == 0:
            logger.info(f"No record in {last_failure} file. Nothing to do, exiting.")
            sys.exit(0)
        logger.info(
            f"Using a file with {len(users)} failed attempts for this config ({last_failure})"
        )
    else:
        all_users: list[User] = []
        for item in settings.source.to:
            all_users.append(
                User(
                    username=parseaddr(item)[0],
                    email=parseaddr(item)[1],
                    lastonline=datetime.now(UTC),
                    status="",
                    uid=-1,
                )
            )
        if settings.source.json_file is not None:
            all_users += read_json(settings.source.json_file)
        users = []
        for user in all_users:
            lastonline = user.lastonline.replace(tzinfo=None)
            if settings.source.online_min <= lastonline <= settings.source.online_max:
                users.append(user)
    ## Process: send emails
    results = await process_users(users)
    ## Collect success
    if len(results.success) > 0:
        logger.info(f"{len(results.success)} mails sent successfully")
        success_file = result_dir / f"{now}.success"
        with open(success_file, "w") as success_fd:
            dump(results.success, success_fd, default=pydantic_encoder)
    ## Collect errors
    error_file = result_dir / f"{now}.err"
    with open(error_file, "w") as errors_fd:
        dump(results.failures, errors_fd, default=pydantic_encoder)
    if len(results.failures) > 0:
        logger.warning(
            f"{len(results.failures)} mails could not be sent. Users written to file {error_file} for retry."
        )
